wb_ <- function(data, minw, nboot, dist_rad, seed = NULL) {

  y <- parse_data(data)

  if (is.null(minw)) {
    minw <- psy_minw(data)
  }

  assert_na(y)
  assert_positive_int(nboot, greater_than = 2)
  assert_positive_int(minw, greater_than = 2)
  stopifnot(is.logical(dist_rad))

  rng_state <- set_rng(seed = seed)

  nc <- ncol(y)
  nr <- nrow(y)

  point <- nr - minw

  adf_crit <- sadf_crit <- gsadf_crit <-
    array(NA, dim = c(nboot, nc),
          dimnames = list(NULL, colnames(y)))

  badf_crit <- bsadf_crit <-
    array(NA, dim = c(point, nboot, nc),
          dimnames = list(NULL, NULL, colnames(y)))

  show_pb <- getOption("exuber.show_progress")
  pb <- set_pb(show_pb, nboot, width = getOption("width") - 15)
  opts <- set_pb_opts(show_pb, pb)

  do_par <- getOption("exuber.parallel")
  if (do_par) {
    cl <- parallel::makeCluster(getOption("exuber.ncores"), type = "PSOCK")
    registerDoSNOW(cl)
    on.exit(parallel::stopCluster(cl))
  }

  `%fun%` <- if (do_par) `%dopar%` else `%do%`

  for (j in 1:nc) {

    dy <- diff(y[, j])
    results <- foreach(
      i = 1:nboot,
      .export = c("rls_gsadf", "unroot"),
      .combine = "cbind",
      .options.snow = opts,
      .inorder = FALSE
    ) %fun% {
      if (show_pb && !do_par)
        setTxtProgressBar(pb, i)
      if (dist_rad) {
        w <- sample(c(-1, 1), nr - 1, replace = TRUE)
      } else {
        w <- rnorm(nr - 1, 0, 1)
      }
      estar <- cumsum(w*dy)
      ystar <- c(0, estar)
      yxmat <- unroot(ystar)
      rls_gsadf(yxmat, min_win = minw)
    }

    if (show_pb)
      cat(paste0(" ", j, "/", nc))

    adf_crit[, j] <- results[point + 1, ]
    sadf_crit[, j] <- results[point + 2, ]
    gsadf_crit[, j] <- results[point + 3, ]

    badf_crit[, , j] <- results[1:point, ]
    bsadf_crit[, , j] <- results[-c(1:(point + 3)), ]
  }

    list(
      adf = adf_crit,
      sadf = sadf_crit,
      gsadf = gsadf_crit,
      badf = badf_crit,
      bsadf = bsadf_crit) %>%
    add_attr(
      seed = rng_state,
      index = attr(y, "index"),
      method = "Wild Bootstrap",
      n = nrow(y),
      minw = minw,
      iter = nboot)


}

#' Wild Bootstrap Critical values
#'
#' \code{wb_cv} performs the Harvey et al. (2016) wild bootstrap re-sampling
#' scheme, which is asymptotically robust to non-stationary volatility, to
#' generate critical values for the recursive unit root tests. \code{wb_dist}
#' computes the distribution.
#'
#' @inheritParams radf
#' @inheritParams mc_cv
#' @param nboot A positive integer indicating the number of bootstraps. Default is 1000 repetitions.
#' @param dist_rad Logical. If \code{TRUE} then  the Rademacher distribution
#' will be used.
#'
#' @return  A list that contains the critical values for ADF, BADF, BSADF and GSADF
#' t-statistics.
#'
#' @details This approach involves applying a wild bootstrap re-sampling scheme
#' to construct the bootstrap analogue of the Phillips et al. (2015) test which
#' is asymptotically robust to non-stationary volatility.
#'
#' @references Harvey, D. I., Leybourne, S. J., Sollis, R., & Taylor, A. M. R.
#' (2016). Tests for explosive financial bubbles in the presence of
#' non-stationary volatility. Journal of Empirical Finance, 38(Part B), 548-574.
#'
#' @references Phillips, P. C. B., Shi, S., & Yu, J. (2015). Testing for
#' Multiple Bubbles: Historical Episodes of Exuberance and Collapse in the
#' S&P 500. International Economic Review, 56(4), 1043-1078.
#'
#' @seealso \code{\link{mc_cv}} for Monte Carlo critical values and
#' \code{\link{sb_cv}} for Sieve Bootstrapped critical values
#'
#' @importFrom doSNOW registerDoSNOW
#' @importFrom parallel detectCores makeCluster stopCluster
#' @importFrom foreach foreach %dopar% %do%
#' @importFrom utils setTxtProgressBar txtProgressBar
#' @importFrom stats quantile rnorm
#' @export
#'
#' @examples
#' \dontrun{
#' # Simulate bubble processes
#' dta <- data.frame(psy1 = sim_psy1(n = 100), psy2 = sim_psy2(n = 100))
#'
#' # Default minimum window
#' wb <- wb_cv(dta)
#'
#' # Change the minimum window and the number of bootstraps
#' wb <- wb_cv(dta, nboot = 1500, minw = 20)
#'
#' # Simulate distribution
#'wb_distr(dta)
#' }
wb_cv <- function(data, minw = NULL, nboot = 1000,
                  dist_rad = FALSE, seed = NULL) {

  results <- wb_(data, minw = minw, nboot = nboot, dist_rad = dist_rad, seed = seed)

  pr <- c(0.9, 0.95, 0.99)

  adf_crit   <- apply(results$adf, 2, quantile, probs = pr) %>% t()
  sadf_crit  <- apply(results$sadf, 2, quantile, probs = pr) %>% t()
  gsadf_crit <- apply(results$gsadf, 2, quantile, probs = pr) %>% t()

  badf_crit  <- apply(results$badf, c(1,3), quantile, probs = pr) %>%
    apply(c(1,3), t)
  bsadf_crit <- apply(results$bsadf, c(1,3), quantile, probs = pr) %>%
    apply(c(1,3), t)

  list(adf_cv = adf_crit,
       sadf_cv = sadf_crit,
       gsadf_cv = gsadf_crit,
       badf_cv = badf_crit,
       bsadf_cv = bsadf_crit) %>%
    inherit_attrs(results) %>%
    add_class("wb_cv", "cv")

}

#' @rdname wb_cv
#' @inheritParams wb_cv
#' @export
wb_distr <- function(data, minw = NULL, nboot = 1000,
                    dist_rad = FALSE, seed = NULL) {

  results <- wb_(data, minw = minw, nboot = nboot, dist_rad = dist_rad, seed = seed)

    list(adf_cv = results$adf,
         sadf_cv = results$sadf,
         gsadf_cv = results$gsadf) %>%
      inherit_attrs(results) %>%
      add_class("wb_distr", "distr")
}

